/*-----------------------------------------------------------------------

File  : clb_pdarrays.c

Author: Stephan Schulz

Contents
 
  Funktions realising the dynamic array type.

Changes

<1> Thu Jul 23 17:40:32 MEST 1998
    New

-----------------------------------------------------------------------*/

#include "clb_pdarrays.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: PDArrayAlloc()
//
//   Return an initialized dynamic array of size init_size.
//
// Global Variables: -
//
// Side Effects    : Memory Operations
//
/----------------------------------------------------------------------*/

PDArray_p PDArrayAlloc(long init_size, long grow)
{
   PDArray_p handle = PDArrayCellAlloc();
   int i;
   
   assert(init_size > 0);
   assert(grow > 0);

   handle->size  = init_size;
   handle->grow  = grow;
   handle->array = SizeMalloc(handle->size*sizeof(IntOrP));
   for(i=0; i<handle->size; i++)
   {
      handle->array[i].p_val = NULL;
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PDArrayFree()
//
//   Free a PDArray. Leaves elements untouched.
//
// Global Variables: -
//
// Side Effects    : Memory Operations
//
/----------------------------------------------------------------------*/

void PDArrayFree(PDArray_p junk)
{
   assert(junk);
   assert(junk->size > 0);
   assert(junk->array);

   SizeFree(junk->array, junk->size*sizeof(IntOrP));
   PDArrayCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: PDArrayElementRef()
//
//   Return a reference to an element in a dynamic array. This
//   reference is only good until the next call to this function! User
//   programs are expected to use this function only extremely rarely
//   and with special care. Use PDArrayElement()/PDArrayAssign()
//   instead. 
//
// Global Variables: -
//
// Side Effects    : May enlarge and move array.
//
/----------------------------------------------------------------------*/

IntOrP* PDArrayElementRef(PDArray_p array, long index)
{
   assert(array);
   assert(index >= 0);
   
   if(!(index < array->size))
   {
      IntOrP *tmp;
      int    old_size, i;
      
      /* Enlarge array */

      old_size = array->size;
      tmp      = array->array;
      array->size = ((index/array->grow)+1)*array->grow;
      array->array = SizeMalloc(array->size * sizeof(IntOrP));
      memcpy(array->array, tmp, old_size*sizeof(IntOrP));
      SizeFree(tmp, old_size * sizeof(IntOrP));
      for(i=old_size; i<array->size; i++)
      {
	 array->array[i].p_val = NULL;
      }      
   }
   return &(array->array[index]);
}


/*-----------------------------------------------------------------------
//
// Function: PDArrayMembers()
//
//   Return number of non-NULL elements in the array.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long PDArrayMembers(PDArray_p array)
{
   long i, res =0;
   
   assert(array);

   for(i=0; i<array->size; i++)
   {
      if(PDArrayElementP(array, i))
      {
	 res++;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: PDArrayFirstUnused()
//
//   Return 1 + the index of the largest element != NULL in array (0
//   if the array is empty).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long PDArrayFirstUnused(PDArray_p array)
{
   long i;
   
   assert(array);

   for(i=array->size; i; i--)
   {
      if(PDArrayElementP(array, i-1))
      {
	 break;
      }
   }
   return i;
}


/*-----------------------------------------------------------------------
//
// Function: PDArrayStore()
//
//   Store the given value after the end of the used part of the
//   array. This is similar to PStackPush() for stacks, but a LOT less
//   efficient. Return value is the index assigned.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long PDArrayStore(PDArray_p array, IntOrP value)
{
   long index;
   
   assert(array);

   index = PDArrayFirstUnused(array);
   PDArrayAssign(array, index, value);

   return index;   
}

/*-----------------------------------------------------------------------
//
// Function: PDArrayStoreP()
//
//   Store the given pointer value after the end of the used part of
/    the array. See PDArrayStore().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long PDArrayStoreP(PDArray_p array, void* value)
{
   IntOrP tmp;
   
   tmp.p_val = value;

   return PDArrayStore(array, tmp);
}


/*-----------------------------------------------------------------------
//
// Function: PDArrayStoreInt()
//
//   Store the given long int value after the end of the used part of
/    the array. See PDArrayStore().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long PDArrayStoreInt(PDArray_p array, long value)
{
   IntOrP tmp;
   
   tmp.i_val = value;

   return PDArrayStore(array, tmp);
}


	
/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


