/*-----------------------------------------------------------------------

File  : cco_forward_contraction.c

Author: Stephan Schulz

Contents
 
  Functions that apply the processed clause sets to simplify or
  eleminate a potential new clause. Extracted from
  cco_proofproc.[ch]. 

Changes

<1> Mon Nov  9 17:46:50 MET 1998
    New

-----------------------------------------------------------------------*/

#include "cco_forward_contraction.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: forward_contract_keep()
//
//   Apply all forward-contracting inferences to clause. Return false
//   it becomes trivial, return true otherwise. Does not delete
//   clause. Subsumed and trivial clauses are counted in the cells
//   pointed to by the 3rd and 4th argument. Provide dummies to avoid
//   this. 
//
// Global Variables: -
//
// Side Effects    : Marks maximals terms and literals in the clause.
//
/----------------------------------------------------------------------*/

static bool forward_contract_keep(ProofState_p state, ProofControl_p
				  control, Clause_p clause, long*
				  subsumed_count, long* trivial_count)
{
   assert(clause);
   assert(state);

   ClauseComputeLINormalform(control->ordering,
			     state->terms, clause,
			     state->demods, FullRewrite);
   
   /* This is the success case! -> There is another one further down. */
   if(ClauseRemoveSuperfluousLiterals(clause))
   {
      return true;
   }

   DEBUGMARK(PP_LOWDETAILS, "Clause is not empty.\n");
   
   if(ClauseIsTautology(clause))
   {
      (*trivial_count)++;
      return false;
   }
   DEBUGMARK(PP_LOWDETAILS, "Clause is not trivial.\n");
   
   /* Now we mark maximal terms... */
   
   ClauseOrientLiterals(control->ordering, clause);

   DEBUGOUT(PP_LOWDETAILS,"# Oriented: ");
   DEBUG(PP_LOWDETAILS,
	 ClausePrint(stdout, clause, true);printf("\n"););
   
   /* Still forward subsumption... */
   
   if(clause->neg_lit_no && 
      ClauseRemoveSubsumedNegLiterals(state->processed_pos_eqns, 
				      clause))
   {
      return true;
   }
   
   ClauseMarkMaximalLiterals(control->ordering, clause);
   
   if((clause->pos_lit_no && 
       UnitClauseSetSubsumesClause(state->processed_pos_eqns,
				   clause)) ||
      (clause->neg_lit_no &&
       UnitClauseSetSubsumesClause(state->processed_neg_units,
				   clause)) ||
      (ClauseLiteralNumber(clause)>1 &&
       ClauseSetSubsumesClause(state->processed_non_units, clause)))
      
   {
      DEBUGMARK(PP_LOWDETAILS, "Clause subsumed!\n");
      (*subsumed_count)++;
      return false;
   }
   return true;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: ForwardContractClause()
//
//   Apply all forward-contracting inferences to clause. Return false
//   and delete the clause if it becomes trivial, return true
//   otherwise. 
//
// Global Variables: -
//
// Side Effects    : Marks maximals terms and literals in the clause.
//
/----------------------------------------------------------------------*/

bool ForwardContractClause(ProofState_p state, ProofControl_p control,
			   Clause_p clause)
{
   bool res;

   assert(clause);
   assert(state);
   
   res = forward_contract_keep(state, control, clause,
			       &(state->proc_forward_subsumed_count),
			       &(state->proc_trivial_count));
   
   if(!res)
   {
      ClauseFree(clause);
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: ForwardContractSet()
//
//   Apply the forward-contracting inferences to all clauses in
//   set. Delete redundant clauses. Return empty clause (if found), 
//   NULL otherwise. The empty clause will be extracted from set,
//   which may not be fully contracted in this case.
//
// Global Variables: -
//
// Side Effects    : Changes clauses and set.
//
/----------------------------------------------------------------------*/

Clause_p ForwardContractSet(ProofState_p state, ProofControl_p
			    control, ClauseSet_p set, long*
			    count_eleminated)
{
   Clause_p handle, next;

   assert(state);
   assert(set);
   assert(!set->demod_index);
   
   handle = set->anchor->succ;
   while(handle != set->anchor)
   {      
      next = handle->succ;
      
      assert(handle);

      if(forward_contract_keep(state, control, handle,
			       count_eleminated, count_eleminated))
      {
	 if(ClauseIsEmpty(handle))
	 {
	    ClauseSetExtractEntry(handle);
	    return handle;
	 }
      }
      else
      {
	 ClauseDetachParents(handle);
	 ClauseSetDeleteEntry(handle);
      }
      handle = next;
   }
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetReweight()
//
//   Re-Evaluate all clauses in set.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetReweight(HCB_p heuristic, ClauseSet_p set)
{
   Clause_p    handle;
   ClauseSet_p tmp_set;

   assert(heuristic);
   assert(set);
   assert(!set->demod_index);
   

   ClauseSetRemoveEvaluations(set);
   tmp_set = ClauseSetAlloc();
   
   while(!ClauseSetEmpty(set))
   {
      ClauseSetInsert(tmp_set, ClauseSetExtractFirst(set));
   }
   while(!ClauseSetEmpty(tmp_set))
   {
      handle = ClauseSetExtractFirst(tmp_set);
      HCBClauseEvaluate(heuristic, handle);
      ClauseSetInsert(set, handle);
   }
   ClauseSetFree(tmp_set);
}




/*-----------------------------------------------------------------------
//
// Function:  ForwardContractSetReweight()
//
//   Apply contracting inferences to all claues in set, then
//   reevaluate them. Return empty clause (if found), NULL
//   otherwise. The empty clause will be extracted from set, which may
//   not be fully contracted in this case. 
//
// Global Variables: -
//
// Side Effects    :  Changes clauses and set.
//
/----------------------------------------------------------------------*/

Clause_p ForwardContractSetReweight(ProofState_p state, ProofControl_p
				    control, ClauseSet_p set, long*
				    count_eleminated)
{
   Clause_p    handle;

   assert(state);
   assert(set);
   assert(!set->demod_index);
   
   handle = ForwardContractSet(state, control, set, count_eleminated);
   
   if(handle)
   {
      return handle;
   }
   ClauseSetReweight(control->heuristic, set);

   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFilterReweigth()
//
//   Remove all tautological clauses from set and reweigth it.
//
// Global Variables: -
//
// Side Effects    : Changes set and *count_eleminated
//
/----------------------------------------------------------------------*/

void ClauseSetFilterReweigth(ProofState_p state, ProofControl_p
			     control, ClauseSet_p set,
			     long* count_eleminated)
{
   *count_eleminated += ClauseSetFilterTautologies(set);
   ClauseSetReweight(control->heuristic, set);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


