/*-----------------------------------------------------------------------

File  : cto_proofcontrol.c

Author: Stephan Schulz

Contents
 
  Basic functions for proof control objects.

Changes

<1> Fri Oct 16 16:20:00 MET DST 1998
    Salvaged from cco_proofstate.h, forked control and state.

-----------------------------------------------------------------------*/

#include "cco_proofcontrol.h"


/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ProofControlFree()
//
//   Allocate an empty, initialized ProofControlCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

ProofControl_p ProofControlAlloc()
{
   ProofControl_p handle = ProofControlCellAlloc();

   handle->ordertype               = NoOrdering;
   handle->to_weight_gen           = WNoMethod;
   handle->to_prec_gen             = PNoMethod;
   handle->to_const_weight         = 1;   
   handle->ordering                = NULL;
   handle->heuristic               = NULL;
   handle->use_indexing            = true;
   handle->forward_demod           = FullRewrite;
   handle->prefer_initial_clauses  = false;
   handle->paramod_with_units_only = false;

   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: ProofControlFree()
//
//   Free a ProofControlCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ProofControlFree(ProofControl_p junk)
{
   assert(junk);
   if(junk->ordering)
   {
      OCBFree(junk->ordering);
   }
   if(junk->heuristic)
   {
      HCBFree(junk->heuristic);
   }


   ProofControlCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ProofControlCreateOrdering()
//
//   Given a proof state, generate a (hopefully suitable) ordering for
//   it. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

OCB_p ProofControlCreateOrdering(ProofState_p state, ProofControl_p control)
{
   OCB_p handle;

   switch(control->ordertype)
   {      
   case LPO:
	 handle = OCBAlloc(LPO, state->signature);
	 TOGeneratePrecedence(handle, state->axioms, PNoMethod);
	 OUTPRINT(1, "# Creating term ordering: LPO selected\n");
	 break;
   case KBO:
	 handle = OCBAlloc(KBO, state->signature);
	 TOGeneratePrecedence(handle, state->axioms, PNoMethod);
	 TOGenerateWeights(handle, state->axioms,
			   control->to_weight_gen,
			   control->to_const_weight);
	 OUTPRINT(1, "# Creating term ordering: KBO selected\n");
	 break;
   case NoOrdering:
	 /* Do some magic...currently default to KBO2! -> Fall through! */
   case KBO2:
	 handle = OCBAlloc(KBO2, state->signature);
	 TOGeneratePrecedence(handle, state->axioms, PNoMethod);
	 TOGenerateWeights(handle, state->axioms,
			   control->to_weight_gen,
			   control->to_const_weight);
	 OUTPRINT(1, "# Creating term ordering: KBO2 selected\n");
	 break;
   case RPO:
	 assert(false && "RPO not yet implemented!");
	 handle = NULL;
	 break;
   default:
	 assert(false);
	 handle = NULL;
	 break;
   }
   /* Select proper literal ordering for Dershowitz-Hack */
   if(control->paramod_with_units_only)      
   {
      handle->neg_pos_uncomparable = true;
   }
   return handle;
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
