/*-----------------------------------------------------------------------

File  : cco_proofproc.c

Author: Stephan Schulz

Contents
 
  Functions realizing the proof procedure.

Changes

<1> Mon Jun  8 11:47:44 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_proofproc.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: remove_unit_subsumed()
//
//   Remove all clauses subsumed by subsumer from set, kill their
//   children. Return number of removed clauses.
//
// Global Variables: -
//
// Side Effects    : Changes set, memory operations.
//
/----------------------------------------------------------------------*/

static long remove_unit_subsumed(Clause_p subsumer, ClauseSet_p set)
{
   Clause_p handle, next;
   long     res = 0;

   assert(ClauseLiteralNumber(subsumer) == 1);

   DEBUGMARK(PP_HIGHDETAILS , "remove_unit_subsumed()...\n");
   next = set->anchor->succ;
   while((handle = ClauseSetFindUnitSubsumedClause(set, next,
						   subsumer)))
   {
      res++;
      DEBUGMARK(PP_HIGHDETAILS, "*\n");
      next = handle->succ;      
      ClauseKillChildren(handle);
      if(set->demod_index)
      {	 
	 ClauseSetIndexedDeleteEntry(handle);      
      }
      else
      {
	 ClauseSetDeleteEntry(handle);  
      }
   }
   DEBUGMARK(PP_HIGHDETAILS, "...remove_ubit_subsumed()\n");

   return res;
}

/*-----------------------------------------------------------------------
//
// Function: remove_subsumed()
//
//   Remove all clauses subsumed by subsumer from set, kill their
//   children. Return number of removed clauses.
//
// Global Variables: -
//
// Side Effects    : Changes set, memory operations.
//
/----------------------------------------------------------------------*/

static long remove_subsumed(Clause_p subsumer, ClauseSet_p set)
{
   Clause_p handle, next;
   long     res = 0;

   assert(ClauseLiteralNumber(subsumer) > 1);

   DEBUGMARK(PP_HIGHDETAILS , "remove_subsumed()...\n");
   next = set->anchor->succ;
   while((handle = ClauseSetFindSubsumedClause(set, next,
					       subsumer)))
   {
      res++;
      DEBUGMARK(PP_HIGHDETAILS, "*\n");
      next = handle->succ;      
      ClauseKillChildren(handle);
      if(set->demod_index)
      {	 
	 ClauseSetIndexedDeleteEntry(handle);      
      }
      else
      {
	 ClauseSetDeleteEntry(handle);  
      }
   }
   DEBUGMARK(PP_HIGHDETAILS, "...remove_subsumed()\n");

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: eleminate_backward_rewritten_clauses()
//
//   Remove all processed clauses rewritable with clause and put
//   them into state->tmp_store.
//
// Global Variables: -
//
// Side Effects    : Changes clause sets
//
/----------------------------------------------------------------------*/

static SysDate
eleminate_backward_rewritten_clauses(ProofState_p
				     state, ProofControl_p control,
				     Clause_p clause)
{
   SysDate date = ClauseSetListGetMaxDate(state->demods,
					  FullRewrite);
   long res = 0, tmp;

   if(ClauseIsDemodulator(clause))
   {
      SysDateInc(&date);
      tmp = RemoveClausesWithRewritableMaxSides(control->ordering,
						state->processed_pos_rules,
						state->tmp_store,
						clause, date);
      /* printf("Removed %ld rules by bw-rewriting.\n", tmp); */
      res += tmp;
      tmp = RemoveClausesWithRewritableMaxSides(control->ordering,
						state->processed_pos_eqns,
						state->tmp_store,
						clause, date);
      /* printf("Removed %ld equations by bw-rewriting.\n", tmp); */
      res += tmp;
      tmp = RemoveClausesWithRewritableMaxSides(control->ordering, 
						state->processed_neg_units,
						state->tmp_store,
						clause, date);
      /* printf("Removed %ld neg unit clauses by bw-rewriting.\n", tmp); */
      res += tmp;
      res += RemoveClausesWithRewritableMaxSides(control->ordering, 
						 state->processed_non_units,
						 state->tmp_store,
						 clause, date);
      state->backward_rewritten_count+=res;
   }
   /* printf("Removed %ld clauses\n", res); */
   return date;
}
   

/*-----------------------------------------------------------------------
//
// Function: eleminate_backward_subsumed_clauses()
//
//   Eleminate subsumed processed clauses, return number of clauses
//   deleted. 
//
// Global Variables: -
//
// Side Effects    : Changes clause sets.
//
/----------------------------------------------------------------------*/


static long eleminate_backward_subsumed_clauses(ProofState_p state,
						Clause_p clause)
{
   long res = 0;

   if(ClauseLiteralNumber(clause) == 1)
   {
      if(clause->pos_lit_no)
      {
	 /* res += remove_subsumed(clause,
	    state->processed_pos_rules); Should be impossible! */
	 res += remove_unit_subsumed(clause, state->processed_pos_eqns);
	 res += remove_unit_subsumed(clause, state->processed_non_units);
      }
      else
      {
	 res += remove_unit_subsumed(clause, state->processed_neg_units);
	 res += remove_unit_subsumed(clause, state->processed_non_units);
      }
   }
   else
   {
      res += remove_subsumed(clause, state->processed_non_units);
   }
   state->backward_subsumed_count+=res;
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: interreduce_clause_sets()
//
//   Rewrite all clause sets in state with
//   state->processed_pos_clauses. Assumes that all maximal sides of
//   maximal literals are irreducible!
//
// Global Variables: -
//
// Side Effects    : Changes clauses in sets.
//
/----------------------------------------------------------------------*/

static void interreduce_clause_sets(ProofState_p state, ProofControl_p
				    control)
{
   ClauseSetComputeLINormalform(control->ordering,
				state->terms,
				state->processed_pos_rules,
				state->demods, FullRewrite);
   ClauseSetComputeLINormalform(control->ordering,
				state->terms,
				state->processed_pos_eqns,
				state->demods, FullRewrite);
   ClauseSetComputeLINormalform(control->ordering,
				state->terms,
				state->processed_neg_units,
				state->demods, FullRewrite);
   ClauseSetComputeLINormalform(control->ordering,
				state->terms,
				state->processed_non_units,
				state->demods, FullRewrite);
}



/*-----------------------------------------------------------------------
//
// Function: insert_new_clauses()
//
//   Rewrite clauses in state->tmp_store, remove superfluous literals,
//   insert them into state->unprocessed. If an empty clause is
//   detected, return it, otherwise return NULL.
//
// Global Variables: -
//
// Side Effects    : As described.
//
/----------------------------------------------------------------------*/

Clause_p insert_new_clauses(ProofState_p state, ProofControl_p control)
{
   Clause_p handle;

   state->generated_count+=state->tmp_store->members;	 
   while((handle = ClauseSetExtractFirst(state->tmp_store)))
   {
      DEBUGMARK(PP_INSERTING_NEW, "insert_new_clauses() - processing: ");
      DEBUG(PP_INSERTING_NEW, ClausePrint(stdout, handle, true););
      DEBUG(PP_INSERTING_NEW, printf("\n"););
      ClauseComputeLINormalform(control->ordering,
				state->terms, handle,
				state->demods, control->forward_demod);
      DEBUGMARK(PP_INSERTING_NEW, "insert_new_clauses() normal form build\n");
      if(ClauseRemoveSuperfluousLiterals(handle))
      {
	 return handle;	 
      }
      DEBUGMARK(PP_INSERTING_NEW, "insert_new_clauses() Literals removed\n");
      if(ClauseIsTautology(handle))
      {
	 assert(!handle->children);
	 ClauseDetachParents(handle);
	 ClauseFree(handle);
      }
      else
      {
	 state->non_trivial_generated_count++;
	 HCBClauseEvaluate(control->heuristic, handle);
	 ClauseSetInsert(state->unprocessed, handle);
      }
   }
   return NULL;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: InitProofState()
//
//   Given a proof state with axioms and a heuristic parameter
//   description, initialize the ProofStateCell, i.e. generate the
//   HCB, the ordering, and evaluate the axioms and put them in the
//   unprocessed list.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void InitProofState(ProofState_p state, ProofControl_p control,
		    HeuristicParms_p h_parms)
{
   Clause_p handle, new;

   OUTPRINT(1, "# Initializing proof state\n");
   
   if(!control->ordering)
   {
      control->ordering = ProofControlCreateOrdering(state, control);
      /* SigPrint(GlobalOut, state->signature); */
      /* OCBDebugPrint(GlobalOut, state->ordering); */
   }
   if(!control->heuristic)
   {
      control->heuristic = HCBCreate(h_parms->heuristic_name,
				     control->ordering, state->signature,
				     state->axioms, h_parms->i_parms,
				     h_parms->d_parms,
				     h_parms->s_parms);
   }
   DEBUGMARK(PP_LOWDETAILS,"InitProofState()...starting to copy\n");
   
   while(!ClauseSetEmpty(state->processed_pos_rules))
   {
      handle = DemodExtract(state->processed_pos_rules->anchor->succ);
      HCBClauseEvaluate(control->heuristic, handle);
      ClauseSetInsert(state->unprocessed, handle);
   }
   while(!ClauseSetEmpty(state->processed_pos_eqns))
   {
      handle = DemodExtract(state->processed_pos_eqns->anchor->succ);
      HCBClauseEvaluate(control->heuristic, handle);
      ClauseSetInsert(state->unprocessed, handle);
   }

   if(control->use_indexing)
   {
      OUTPRINT(1, "# Using perfect discrimination trees\n");
      state->processed_pos_rules->demod_index =
	 PDTreeAlloc(state->signature->f_count+1);
      state->processed_pos_eqns->demod_index =
	 PDTreeAlloc(state->signature->f_count+1);
   }

   while(!ClauseSetEmpty(state->processed_non_units))
   {
      handle = state->processed_non_units->anchor->succ;
      ClauseSetExtractEntry(handle);
      HCBClauseEvaluate(control->heuristic, handle);
      ClauseSetInsert(state->unprocessed, handle);
   }
   
   for(handle = state->axioms->anchor->succ;
       handle != state->axioms->anchor; handle = handle->succ)
   {
      new = ClauseCopy(handle, state->terms);
      HCBClauseEvaluate(control->heuristic, new);
      if(control->prefer_initial_clauses)
      {
	 EvalListSetPriority(new->evaluations, PrioPrefer);
      }
      ClauseSetInsert(state->unprocessed, new);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ProcessClause()
//
//   Select an unprocessed clause, process it. Return pointer to empty
//   clause if it can be derived, NULL otherwise.
//
// Global Variables: -
//
// Side Effects    : Everything ;-)
//
/----------------------------------------------------------------------*/

Clause_p ProcessClause(ProofState_p state, ProofControl_p control)
{
   Clause_p clause, tmp_copy, empty;
   SysDate  clausedate;

   VERBOUT2("Processing clause\n");
   
   clause = control->heuristic->hcb_select(control->heuristic,
					 state->unprocessed);
   
   if(OutputLevel)
   {
      putc('#', GlobalOut);
   }
   
   assert(clause);
   DEBUGOUT(PP_LOWDETAILS, "# *** Processing Clause: ");
   DEBUG(PP_LOWDETAILS, ClausePrint(stdout, clause, true);); 
   DEBUGOUT(PP_LOWDETAILS, "\n");
   DEBUGOUT(PP_HIGHDETAILS, "Processed clauses:\n");
   DEBUG(PP_HIGHDETAILS, ClauseSetPrint(stdout,
					state->processed_pos_rules,
					true););
   DEBUG(PP_HIGHDETAILS, ClauseSetPrint(stdout,
					state->processed_pos_eqns,
					true););    
   DEBUG(PP_HIGHDETAILS, ClauseSetPrint(stdout,
					state->processed_neg_units,
					true);); 
   DEBUG(PP_HIGHDETAILS, ClauseSetPrint(stdout,
					state->processed_non_units,
					true);); 
   
   state->processed_count++;
   
   ClauseSetExtractEntry(clause);
   ClauseDetachParents(clause);
   ClauseRemoveEvaluations(clause);

   if(!ForwardContractClause(state, control, clause))
   {
      return NULL;
   }

   if(ClauseIsEmpty(clause))
   {
      return clause;
   }

   if(OutputLevel)
   {
      putc('\n', GlobalOut);
      ClausePrint(GlobalOut, clause, true);
      putc('\n', GlobalOut);
   }
   
   state->proc_non_trivial_count++;
   
   /* Now on to backward subsumption. */   

   clausedate = eleminate_backward_rewritten_clauses(state, control,
						     clause);
   eleminate_backward_subsumed_clauses(state, clause);
   
   tmp_copy = ClauseCopy(clause, state->tmp_terms);      

   ClauseMarkRestrictedTerms(clause);
   
   DEBUGOUT(PP_LOWDETAILS,"# Restricted: ");
   DEBUG(PP_LOWDETAILS,
	 ClausePrint(stdout, clause, true);printf("\n"););   

   if(ClauseIsDemodulator(clause))
   {
      assert(clause->neg_lit_no == 0);
      clause->date = clausedate;
      if(EqnIsOriented(clause->literals))
      {
	 state->processed_pos_rules->date = clausedate;
	 DemodInsert(state->processed_pos_rules, clause);
      }
      else
      {
	 state->processed_pos_eqns->date = clausedate;
	 DemodInsert(state->processed_pos_eqns, clause);
      }
      interreduce_clause_sets(state, control);
   }
   else if(ClauseLiteralNumber(clause) == 1)
   {
      assert(clause->neg_lit_no == 1);
      DemodInsert(state->processed_neg_units, clause);
   }
   else
   {
      DemodInsert(state->processed_non_units, clause);
   }
      
   DEBUGMARK(PP_LOWDETAILS, "...starting generating inferences...\n");

   state->factor_count+=
      ComputeAllEqualityFactors(state->terms, control->ordering, clause,
			       state->tmp_store, state->freshvars);
   DEBUGMARK(PP_HIGHDETAILS, "...starting equality resolution...\n");
   state->resolv_count+=
      ComputeAllEqnResolvents(state->terms, clause, state->tmp_store,
			      state->freshvars);
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (pos_rules)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->terms, control->ordering,
			      tmp_copy, clause,
			      state->processed_pos_rules,
			      state->tmp_store, state->freshvars,
			      control->paramod_with_units_only); 
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (pos_eqns)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->terms, control->ordering,
			      tmp_copy, clause,
			      state->processed_pos_eqns,
			      state->tmp_store, state->freshvars,
			      control->paramod_with_units_only); 
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (neg_units)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->terms, control->ordering,
			      tmp_copy, clause,
			      state->processed_neg_units,
			      state->tmp_store, state->freshvars,
			      control->paramod_with_units_only);
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (non_units)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->terms, control->ordering,
			      tmp_copy, clause,
			      state->processed_non_units, 
			      state->tmp_store, state->freshvars,
			      control->paramod_with_units_only);
   DEBUGMARK(PP_LOWDETAILS, "...completed generation...\n");
   ClauseFree(tmp_copy);
   
   if((empty = insert_new_clauses(state, control)))
   {
      return empty;
   }  
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function:  Saturate()
//
//   Process clauses until either the empty clause has been derived, a
//   specified number of clauses has been processed, or the clause set
//   is saturated. Return emoty clause (if found) or NULL. 
//
// Global Variables: -
//
// Side Effects    : Modifies state.
//
/----------------------------------------------------------------------*/

Clause_p Saturate(ProofState_p state, ProofControl_p control, long
		  step_limit, long proc_limit, long unproc_limit, long
		  total_limit)
{
   Clause_p unsatisfiable = NULL;
   long count = 0;
   /* long limit = 5000; */

   while(state->unprocessed->members && 
	 step_limit>=count && 
	 proc_limit>=(state->processed_pos_rules->members +
		      state->processed_pos_eqns->members +
		      state->processed_neg_units->members +
		      state->processed_non_units->members) &&
	 unproc_limit>=state->unprocessed->members &&
	 total_limit>=(state->processed_pos_rules->members +
		       state->processed_pos_eqns->members +
		       state->processed_neg_units->members +
		       state->processed_non_units->members+
		       state->unprocessed->members))
   {     
      count++;
      unsatisfiable = ProcessClause(state, control);
      if(unsatisfiable)
      {
	 break;
      }
#ifdef NEVER_DEFINED
      if((state->proc_non_trivial_count % limit) == 0)
      {
	 unsatisfiable = 
	    ForwardContractSetReweight(state, control,
				       state->unprocessed, 
				       &(state->other_redundant_count));
	 ClauseSetFilterReweigth(state, control, state->unprocessed,
	       &(state->other_redundant_count));
      }
#endif
   }
   return unsatisfiable;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/



