/*-----------------------------------------------------------------------

File  : che_to_weightgen.c

Author: Stephan Schulz

Contents
 
  Functions implementing several simple weight generation schemes for
  the KBO.

Changes

<1> Fri Sep 25 02:49:11 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_to_weightgen.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

char* TOWeightGenNames[]=
{
   "none",            /* WNoMethod */
   "constant",        /* WConstantWeight */
   "maximal0",        /* WSelectMaximal */
   "unary0",          /* WSelectArity1 */
   "single_unary0",   /* WCondSelectArity1 */
   "first_unary0",    /* WSelectFirstArity1 */
   "freq_unary0",     /* WSelectFreqArity1 */
   "arity",           /* WArityWeight */
   "invarity",        /* WAritySqWeight */
   "aritysquared",    /* WInvArityWeight */
   "invaritysquared", /* WInvSqArityWeight */
   "precedence",      /* WPrecedence */
   "invprecedence",   /* WPrecedenceInv */
   NULL
};


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: generate_constant_weights()
//
//   Assign the constant W_DEFAULT_WEIGHT to all smbols. 
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_constant_weights(OCB_p ocb)
{
   FunCode i;

   for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
   {
      *OCBFunWeightPos(ocb, i) = W_DEFAULT_WEIGHT;
   }   
}

/*-----------------------------------------------------------------------
//
// Function: generate_selmax_weights()
//
//   Assign weight W_DEFAULT_WEIGHT to all symbols except the maximal
//   ones, which get weight 0. Not necessarily a reduction ordering!
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_selmax_weights(OCB_p ocb)
{
   FunCode i,j;
   bool    max;
   assert(ocb->precedence);
   
   for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
   {
      max = true;

      for(j=1; j<=ocb->sig_size; j++)
      {
	 if(OCBFunCompare(ocb, i, j) == to_lesser)
	 {
	    max = false;
	    break;
	 }
      }   
      *OCBFunWeightPos(ocb, i) = max?0:W_DEFAULT_WEIGHT;
   }   
}

/*-----------------------------------------------------------------------
//
// Function: generate_selarity1_weights()
//
//   Assign weight 0 to all symbols of arity 1, W_DEFAULT_WEIGHT to
//   all other symbols. 
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_selarity1_weights(OCB_p ocb)
{
   FunCode i;

   for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
   {
      *OCBFunWeightPos(ocb, i) = SigFindArity(ocb->sig, i)
	 == 1 ? 0:W_DEFAULT_WEIGHT;
   }   
}


/*-----------------------------------------------------------------------
//
// Function: generate_condselarity1_weights()
//
//   If there is a unique symbol of arity 1, give it weight 0,
//   otherwise give weight W_DEFAULT_WEIGHT.
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_condselarity1_weights(OCB_p ocb)
{
   FunCode i;
   int count = 0;

   for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
   {
      if(SigFindArity(ocb->sig, i)==1)
      {
	 count++;
      }
   }
   if(count>1)
   {
      generate_constant_weights(ocb);
   }
   else
   {
      generate_selarity1_weights(ocb);
   }
}


/*-----------------------------------------------------------------------
//
// Function: generate_firstarity1_weights()
//
//   Give weight 0 to the first symbol with arity 1.
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_firstarity1_weights(OCB_p ocb)
{
  FunCode i;
  
  generate_constant_weights(ocb);
  
  for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
    { 
      if(SigFindArity(ocb->sig, i)==1)
	{
	  *OCBFunWeightPos(ocb, i) = 0;
	  break;
	}
    }
}




/*-----------------------------------------------------------------------
//
// Function: generate_selectfreqarity1_weights()
//
//   Give weight 0 to the unary function symbol which occurs most
//   often in the axiom, W_DEFAULT_WEIGHT to all other sybols.
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_selectfreqarity1_weights(OCB_p ocb, ClauseSet_p
					      axioms)
{
   FunCode sel = 0, i;
   long    max = 0;
   long    *dist_array;
   
   generate_constant_weights(ocb);

   dist_array = SizeMalloc((ocb->sig->f_count+1) * sizeof(long));
   for(i=0; i<=ocb->sig->f_count; i++)
   {
      dist_array[i] = 0;
   }      
   ClauseSetAddSymbolDistribution(axioms, dist_array);

   for(i=1; i<=ocb->sig->f_count; i++)
   {
      if((SigFindArity(ocb->sig, i)==1) && (dist_array[i] > max))
      {
	 max = dist_array[i];
	 sel = i;
      }
   }
   SizeFree(dist_array, (ocb->sig->f_count+1) * sizeof(long)); 
      
   if(sel)
   {
      *OCBFunWeightPos(ocb, sel) = 0;
   }
}


/*-----------------------------------------------------------------------
//
// Function: generate_arity_weights()
//
//   Generate arity-based weights for function symbols.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void generate_arity_weights(OCB_p ocb, TOWeightGenMethod method)
{
   FunCode i;
   int arity, weight = W_DEFAULT_WEIGHT, maxarity;
   
   assert((method == WArityWeight)||(method == WAritySqWeight)||
	  (method == WInvArityWeight)||(method == WInvAritySqWeight));

   maxarity = SigFindMaxUsedArity(ocb->sig);	    
   for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
   { 
      arity = SigFindArity(ocb->sig, i);
      switch(method)
      {
      case WArityWeight:
	    weight = arity+1;
	    break;
      case WAritySqWeight:
	    weight = arity*arity+1;
	    break;
      case WInvArityWeight:
	    weight = maxarity-arity+1;
	    break;
      case WInvAritySqWeight:
	    weight = (maxarity*maxarity)-(arity*arity)+1;
	    break;
      default:
	    assert(false);
	    break;
      }
      *OCBFunWeightPos(ocb, i) = weight*W_DEFAULT_WEIGHT;
   }
}


/*-----------------------------------------------------------------------
//
// Function: generate_precedence_weights()
//
//    Weight(f) = |{g|g<f}|, i.e. weight is number of smaller function
//    symbols in the signature (+1). 
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_precedence_weights(OCB_p ocb)
{
  FunCode i, j;
  int     weight;
  
  assert(ocb->precedence);
  
  for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
  { 
     weight = 1;
     for(j=SIG_TRUE_CODE+1; j<=ocb->sig_size; j++)
     {
	if(OCBFunCompare(ocb, i,j)== to_greater)
	{
	   weight++;
	}
     }
     *OCBFunWeightPos(ocb, i) = weight*W_DEFAULT_WEIGHT;
  }
}


/*-----------------------------------------------------------------------
//
// Function: generate_invprecedence_weights()
//
//    Weight(f) = |{g|g>f}|, i.e. weight is number of bigger function
//    symbols in the signature (+1). 
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

static void generate_invprecedence_weights(OCB_p ocb)
{
  FunCode i, j;
  int     weight;
  
  assert(ocb->precedence);
  
  for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
  { 
     weight = 1;
     for(j=SIG_TRUE_CODE+1; j<=ocb->sig_size; j++)
     {
	if(OCBFunCompare(ocb, i,j)== to_lesser)
	{
	   weight++;
	}
     }
     *OCBFunWeightPos(ocb, i) = weight*W_DEFAULT_WEIGHT;
  }
}




/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TOTranslateWeightGenMethod()
//
//   Given a string, return the corresponding TOWeightGenMethod
//   token. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/


TOWeightGenMethod TOTranslateWeightGenMethod(char* name)
{
   TOWeightGenMethod method;
   
   method = StringIndex(name, TOWeightGenNames);

   if(method == -1)
   {
      method = WNoMethod;
   }
   return method;
}


/*-----------------------------------------------------------------------
//
// Function: TOGenerateWeights()
//
//   Given a pre-initialized OCB, assign weights to the function
//   symbols. Some methods require a precedence, some require the
//   axioms. 
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

void TOGenerateWeights(OCB_p ocb, ClauseSet_p axioms,
		       TOWeightGenMethod method, long const_weight)
{
   assert(ocb);
   assert(ocb->weights);
   assert(ocb->sig);

   *OCBFunWeightPos(ocb,SIG_BOTTOM_CODE) = 0;
   *OCBFunWeightPos(ocb, SIG_TRUE_CODE) = 1;

   switch(method)
   {
   case WConstantWeight:
	 generate_constant_weights(ocb);
	 break;
   case WSelectMaximal:
	 generate_selmax_weights(ocb);
	 break;
   case WSelectArity1:
	 generate_selarity1_weights(ocb);
	 break;
   case WCondSelectArity1:
	 generate_condselarity1_weights(ocb);
	 break;
   case WSelectFirstArity1:
	 generate_firstarity1_weights(ocb);	 
	 break;
   case WSelectFreqArity1:
	 generate_selectfreqarity1_weights(ocb, axioms);
	 break;
   case WArityWeight:
   case WAritySqWeight:
   case WInvArityWeight:
   case WInvAritySqWeight:
	 generate_arity_weights(ocb, method);
	 break;
   case WPrecedence:
	 generate_precedence_weights(ocb);
	 break;
   case WPrecedenceInv:
	 generate_invprecedence_weights(ocb);
	 break;
   case WNoMethod:
	 TOGenerateDefaultWeights(ocb, ocb->sig);
	 break;
   default:
	 assert(false && "Weight generation method unimplemented");
	 break;
   }
   if(const_weight != WConstNoSpecialWeight)
   {
      FunCode i;
      
      for(i=SIG_TRUE_CODE+1; i<=ocb->sig_size; i++)
      {
	 if(SigFindArity(ocb->sig, i)==0)
	 {
	    *OCBFunWeightPos(ocb, i) = const_weight;
	 }
      }
   }
   *OCBFunWeightPos(ocb, SIG_TRUE_CODE) = ocb->var_weight;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


