/*-----------------------------------------------------------------------

File  : eprover.c

Author: Stephan Schulz

Contents
 
  Main program for the E equational theorem prover.

Changes

<1> Tue Jun  9 01:32:15 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include <cio_commandline.h>
#include <cio_output.h>
#include <cco_proofproc.h>
#include <cco_signals.h>

/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

#define VERSION "0.24 Yunnan"

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERSION,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_PRINT_SATURATED,
   OPT_FILTER_SATURATED,
   OPT_PRINT_PID,
   OPT_SILENT,
   OPT_OUTPUTLEVEL,
   OPT_MEM_LIMIT,
   OPT_CPU_LIMIT,
   OPT_RUSAGE_INFO,
   OPT_STEP_LIMIT,
   OPT_PROC_LIMIT,
   OPT_UNPROC_LIMIT,
   OPT_TOTAL_LIMIT,
   OPT_DISCOUNT_VARS,
   OPT_NO_INFIX,
   OPT_FULL_EQ_REP,
   OPT_NO_NEG_PARAMOD,
   OPT_NO_NONEQ_PARAMOD,
   OPT_PARAMOD_WITH_UNITS_ONLY,
   OPT_PREFER_INITIAL_CLAUSES,
   OPT_ORDERING,
   OPT_TO_WEIGHTGEN,
   OPT_TO_PRECGEN,
   OPT_TO_CONSTWEIGHT,
   OPT_INDEXING,   
   OPT_RULES_GENERAL,
   OPT_FORWARD_DEMOD,
   OPT_STRONGSUBSUMPTION,
   OPT_HEURISTIC,
   OPT_HEU_INT,
   OPT_HEU_FLOAT,
   OPT_HEU_STRING
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},

   {OPT_VERSION,
    '\0', "version",
    NoArg, NULL,
    "Print the version number of the prover."},

   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program. This differs "
    "from the output level (below) in that technical information is "
    "printed to stderr, while the output level determines which "
    "logical manipulations of the clauses are printed to stdout."},

   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},

   {OPT_SILENT,
    's', "silent",
    NoArg, NULL,
    "Equivalent to --output-level=0."},

   {OPT_OUTPUTLEVEL,
    'l', "output-level",
    ReqArg, NULL,
    "Select an output level, greater values imply more verbose "
    "output. Level 0 produces "
    "nearly no output, level 1 will output each clause as it is "
    "processed, level 2 will output generating inferences (not yet "
    "implemented)  and level three will give a full protocol "
    "including rewrite steps."},
   
   {OPT_PRINT_SATURATED,
    'S', "print-saturated",
    OptArg, "true",
    "Print the (semi-) saturated clause sets after terminating the "
    "saturation process."},

    {OPT_FILTER_SATURATED,
    'S', "filter-saturated",
    OptArg, "true",
    "Filter the  (semi-) saturated clause sets after terminating the "
    "saturation process, by forward-contracting them."},

   {OPT_PRINT_PID,
    '\0', "print-pid",
    NoArg, NULL,
    "Print the process id of the prover as a comment after option "
    "processing."},

   {OPT_MEM_LIMIT,
    'm', "memory-limit",
    OptArg, "64",
    "Limit the memory the prover may use. The optional argument is "
    "the allowed amount of memory in MB. This option may not work "
    "everywhere, due to broken and/or strange behaviour of setrlimit() "
    "in some UNIX implementations. It does work under all tested "
    "versions of Solaris and GNU/Linux."},   

   {OPT_CPU_LIMIT,
    '\0', "cpu-limit",
    OptArg, "300",
    "Limit the cpu time limit. The optional argument is the CPU time "
    "in seconds. This option may not work "
    "everywhere, due to broken and/or strange behaviour of setrlimit() "
    "in some UNIX implementations. It does work under all tested "
    "versions of Solaris and GNU/Linux."}, 

   {OPT_RUSAGE_INFO,
    'R', "resources-info",
    OptArg, "true",
    "Give some information about the resources used by the prover. "
    "You will usually get CPU time information. On systems returning "
    "more information with the rusage() system call, you will also "
    "get information about memory consumption."},

   {OPT_STEP_LIMIT,
    'C', "processed-clauses-limit",
    ReqArg, NULL,
    "Set the maximal number of clauses to process."},
   
   {OPT_PROC_LIMIT,
    'P', "processed-set-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of processed clauses."},

   {OPT_UNPROC_LIMIT,
    'U', "unprocessed-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of unprocessed clauses."},
   
   {OPT_TOTAL_LIMIT,
    'T', "total-clause-set-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of all clauses."},

   {OPT_DISCOUNT_VARS,
    'd', "discount-vars",
    OptArg, "true",
    "Read and write PCL terms, i.e. write constants as functions of "
    "arity 0 ('a()') and treat all other identifiers as variables. "
    "otherwise use PROLOG terms (variables start with a capital "
    "letter or '_', everything else is a constant)."},

   {OPT_NO_INFIX,
    'n', "eqn-no-infix",
    OptArg, "true",
    "Print equations in prefix notation equal(x,y)."},

   {OPT_FULL_EQ_REP,
    'e', "full-equational-rep",
    OptArg, "true",
    "Print all literals as equations, even non-equational ones."},

   {OPT_NO_NEG_PARAMOD,
    'N', "no-paramod-into-negatives",
    OptArg, "true",
    "Refrain from paramodulating into negative literals."},
   
   {OPT_NO_NONEQ_PARAMOD,
    'E', "paramod-equational-literals-only",
    OptArg, "true",
    "Don't paramodulate from or into non-equational literals."},

   {OPT_PARAMOD_WITH_UNITS_ONLY,
    'W', "paramod-with-units-only",
    OptArg, "true",
    "Paramodulate with unit-clauses only (but into arbitrary "
    "clauses). This should be complete and fairly efficient for horn "
    "problems. Also modifies literal ordering: Negative and positive "
    "literals are made uncomparable."},

   {OPT_PREFER_INITIAL_CLAUSES,
    '\0', "prefer-initial-clauses",
    OptArg, "true",
    "Always process all initial clauses first."},

   {OPT_HEURISTIC,
    'x', "expert-heuristic",
    ReqArg, NULL,
    "Select one of the heuristics. Currently available: Standard, "
    "Clauseweight, FIFOWeight."}, 

   {OPT_ORDERING,
    't', "term-ordering",
    ReqArg, NULL,
    "Select an ordering type (currently LPO, KBO or KBO1)."},      

   {OPT_TO_WEIGHTGEN,
    'w', "order-weight-generation",
    ReqArg, NULL,
    "Select a method for the generation of weights for use with the "
    "term ordering. Run 'eprover -w none' for a list of options."},

   {OPT_TO_PRECGEN,
    'G', "order-precedence-generation",
    ReqArg, NULL,
    "Select a method for the generation of a precedence for use with "
    "the term ordering. Run 'eprover -G none' for a list of "
    "options."}, 

   {OPT_TO_CONSTWEIGHT,
    'c', "order-constant-weight", 
    ReqArg, NULL,
    "Set a special weight > 0 for constants in the term ordering."
    "By default, constants are treated like other function symbols."
   },
   
   {OPT_INDEXING,
    'p', "no-pdt-indexing",
    NoArg, NULL,
    "Use conventional matching for forward rewriting and subsumption "
    "(default is to use perfect discrimination trees)."}, 

   {OPT_RULES_GENERAL,
    'g', "prefer-general-demodulators",
    NoArg, NULL,
    "If using perfect discrimination tree indexing, prefer general "
    "demodulators. By default, use specialized demodulators."},
   
   {OPT_FORWARD_DEMOD,
    'F', "forward_demod_level",
    ReqArg, NULL,
    "Set the desired level rewriting of unprocessed clauses. A value "
    "of 0 means no rewriting, 1 indicates to use rules (orientable "
    "equations) only, 2 indicates full rewriting with rules and "
    "instances of unorientable equations. Default is 2."},

   {OPT_STRONGSUBSUMPTION,
    'u', "strong-forward-subsumption",
    NoArg, NULL,
    "Try multiple positions and unit-equations to try to subsume a "
    "single new clause. Default is to search for a single position."},

   {OPT_HEU_INT,
    'i', "heuristic-int-arg",
    ReqArg, NULL,
    "Give an integer argument to the selected heuristic. This option "
    "can be given multiple times, order is significant."},

   {OPT_HEU_FLOAT,
    'f', "heuristic-float-arg",
    ReqArg, NULL,
   "Give a float argument to the selected heuristic. This option "
    "can be given multiple times, order is significant."},

   {OPT_HEU_STRING,
    's', "heuristic-string-arg",
    ReqArg, NULL,
   "Give a string argument to the selected heuristic. This option "
    "can be given multiple times, order is significant."},

   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

TermOrdering      ordertype = NoOrdering;
char              *outname = NULL;
HeuristicParms_p  h_parms;
bool              indexing = true,
                  print_sat = false,
                  filter_sat = false,
                  print_rusage = false,
                  paramod_with_units_only = false,
                  prefer_initial_clauses = false;
bool              print_pid = false;
long              step_limit = LONG_MAX, 
                  proc_limit = LONG_MAX,
                  unproc_limit = LONG_MAX, 
                  total_limit = LONG_MAX,
                  to_constant_weight = -1;
TOWeightGenMethod weight_gen = WNoMethod;
TOPrecGenMethod   prec_gen = PNoMethod;
RewriteLevel      forward_demod = FullRewrite;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   CLState_p        state;
   Scanner_p        in;    
   ProofState_p     proofstate;
   ProofControl_p   proofcontrol;
   int              i;
   Clause_p         success, filter_success;

   assert(argv[0]);
   InitError(argv[0]);

   if (signal(SIGXCPU, ESignalHandler) == SIG_ERR)
   {
      TmpErrno = errno;
      SysError("Unable to set up signal handler", SYS_ERROR);
   }
   
   DemodOrderCriterion = HCBLIFOCreate(NULL, NULL, NULL, NULL, NULL,
				       NULL);

   h_parms = HeuristicParmsAlloc();
   h_parms->heuristic_name = "Clauseweight";

   state = process_options(argc, argv);

   GlobalOut = OutOpen(outname);

   if(print_pid)
   {
      fprintf(GlobalOut, "# Pid: %d\n", (int)getpid());
      fflush(GlobalOut);
   }

   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }
   
   for(i=0; state->argv[i]; i++)
   {
      in = CreateScanner(state->argv[i] , true, NULL);
      
      proofstate = ProofStateAlloc();
      ClauseSetParseList(in, proofstate->axioms,
			 proofstate->original_terms);

      proofcontrol = ProofControlAlloc();
      proofcontrol->paramod_with_units_only = paramod_with_units_only;
      proofcontrol->prefer_initial_clauses = prefer_initial_clauses;
      proofcontrol->ordertype = ordertype;
      proofcontrol->to_weight_gen = weight_gen;
      proofcontrol->to_prec_gen = prec_gen;
      proofcontrol->to_const_weight = to_constant_weight;
      proofcontrol->forward_demod = forward_demod;      
      
      VERBOUT2("Specification read\n");
      InitProofState(proofstate, proofcontrol, h_parms);
      VERBOUT2("Prover state initialized\n");
      success = Saturate(proofstate, proofcontrol, step_limit,
			 proc_limit, unproc_limit, total_limit);
      
      if(filter_sat)
      {
	 filter_success = ForwardContractSet(proofstate, proofcontrol,
					     proofstate->unprocessed,
					     &(proofstate->proc_trivial_count));
	 if(success)
	 {
	    success = filter_success;
	 }
      }
      
      if(success)
      {
	 fprintf(GlobalOut, "\n# Proof found!\n");
	 ClauseFree(success);
      }
      else
      {
	 fprintf(GlobalOut, "\n# No proof found!\n");
      }
      if(print_sat)
      {
	 ProofStatePrint(GlobalOut, proofstate);
	 fprintf(GlobalOut, "\n");
      }	 
      fflush(GlobalOut);
      
      if(OutputLevel)
      {
	 ProofStateStatisticsPrint(GlobalOut, proofstate);
      }
      ProofControlFree(proofcontrol);
      ProofStateFree(proofstate);
      DestroyScanner(in);
   }

   CLStateFree(state);

   HeuristicParmsFree(h_parms);
   HCBFree(DemodOrderCriterion);

   if(print_rusage)
   {
      struct rusage usage;
      
      if(getrusage(RUSAGE_SELF, &usage))
      {
	 TmpErrno = errno;
	 SysError("Unable to get resource usage information",
		  SYS_ERROR);
      }
      fprintf(GlobalOut, 
	      "\n# -------------------------------------------------\n");
      fprintf(GlobalOut, 
	      "# User time                : %.3f s\n",
	      (usage.ru_utime.tv_sec)+(usage.ru_utime.tv_usec)/1000000.0);
      fprintf(GlobalOut, 
	      "# System time              : %.3f s\n",
	      (usage.ru_stime.tv_sec)+(usage.ru_stime.tv_usec)/1000000.0);
      fprintf(GlobalOut, 
	      "# Total time               : %.3f s\n",
	      (usage.ru_utime.tv_sec+usage.ru_stime.tv_sec)+
	      ((usage.ru_utime.tv_usec+usage.ru_stime.tv_usec)/1000000.0));
      fprintf(GlobalOut, 
	      "# Maximum resident set size: %ld pages\n",
	      usage.ru_maxrss);
   }
   fflush(GlobalOut);
   OutClose(GlobalOut);
   
#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif
   
   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//                   TBPrintInternalInfo 
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   struct rlimit     limit = {RLIM_INFINITY, RLIM_INFINITY};
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	    Verbose = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_HELP: 
	    print_help(stdout);
	    exit(NO_ERROR);
	    break;
      case OPT_VERSION:
	    printf("E " VERSION "\n");
	    exit(NO_ERROR);
	    break;
      case OPT_OUTPUT:
	    outname = arg;
	    break;
      case OPT_SILENT:
	    OutputLevel = 0;
	    break;
      case OPT_OUTPUTLEVEL:
	    OutputLevel = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_PRINT_SATURATED:
	    print_sat = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_FILTER_SATURATED:
	    filter_sat = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_PRINT_PID:
	    print_pid = true;
	    break;
      case OPT_MEM_LIMIT:
	    limit.rlim_cur = MEGA*CLStateGetIntArg(handle, arg);
	    if(setrlimit(RLIMIT_DATA, &limit))
	    {
	       TmpErrno = errno;
	       SysError("Unable to set memory limit", SYS_ERROR);
	    }
	    break;
      case OPT_CPU_LIMIT:
	    limit.rlim_cur = CLStateGetIntArg(handle, arg);
	    if(setrlimit(RLIMIT_CPU, &limit))
	    {
	       TmpErrno = errno;
	       SysError("Unable to set cpu time limit", SYS_ERROR);
	    }
	    break;
      case OPT_RUSAGE_INFO:
	    print_rusage = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_STEP_LIMIT:
	    step_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_PROC_LIMIT:
	    proc_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_UNPROC_LIMIT:
	    unproc_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_TOTAL_LIMIT:
	    total_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_DISCOUNT_VARS:
	    TermPrologVars = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_NO_INFIX:
	    EqnUseInfix = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_FULL_EQ_REP:
	    EqnFullEquationalRep = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_NO_NEG_PARAMOD:
	    ParamodOverlapIntoNegativeLiterals = !CLStateGetBoolArg(handle, arg);
	    break;	    
      case OPT_NO_NONEQ_PARAMOD:
	    ParamodOverlapNonEqLiterals = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_PARAMOD_WITH_UNITS_ONLY:
	    paramod_with_units_only = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_PREFER_INITIAL_CLAUSES:
	    prefer_initial_clauses = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_HEURISTIC:
	    h_parms->heuristic_name = arg;
	    break;
      case OPT_ORDERING:
	    if(strcmp(arg, "LPO")==0)
	    {
	       ordertype = LPO;
	    }
	    else if(strcmp(arg, "KBO1")==0)
	    {
	       ordertype = KBO;
	    }
	    else if(strcmp(arg, "KBO")==0)
	    {
	       ordertype = KBO2;
	    }
	    else
	    {
	       Error("Option -t (--term-ordering) requires LPO, KBO or KBO1"
		     " as an argument", USAGE_ERROR);
	    }
	    break;
      case OPT_TO_WEIGHTGEN:
	    weight_gen = TOTranslateWeightGenMethod(arg);
	    if(!weight_gen)
	    {
	       DStr_p err = DStrAlloc();
	       DStrAppendStr(err, 
			     "Wrong argument to option -w "
			     "(--order-weight-generation). Possible "
			     "values: "); 
	       DStrAppendStrArray(err, TOWeightGenNames, ", ");
	       Error(DStrView(err), USAGE_ERROR);
	       DStrFree(err);
	    }		      
	    break;
      case OPT_TO_PRECGEN:
	    prec_gen = TOTranslatePrecGenMethod(arg);
	    if(!prec_gen)
	    {
	       DStr_p err = DStrAlloc();
	       DStrAppendStr(err, 
			     "Wrong argument to option -G "
			     "(--order-precedence-generation). Possible "
			     "values: "); 
	       DStrAppendStrArray(err, TOPrecGenNames, ", ");
	       Error(DStrView(err), USAGE_ERROR);
	       DStrFree(err);
	    }		      
	    break;
      case OPT_TO_CONSTWEIGHT:
	    to_constant_weight = CLStateGetIntArg(handle, arg); 
	    if(to_constant_weight<=0)
	    {
	       Error("Argument to option -c (--order-constant-weight) "
		     "has to be > 0", USAGE_ERROR);
	    }
	    break;
      case OPT_INDEXING:
	    indexing = false;
	    break;
      case OPT_RULES_GENERAL:
	    PDTreePreferGeneral = true;
	    break;
      case OPT_FORWARD_DEMOD:
	    forward_demod = CLStateGetIntArg(handle, arg);
	    if(forward_demod > 2 || forward_demod < 0)
	    {
	       Error("Option -F (--forward_demod_level) requires "
		     "argument from {0..2}", USAGE_ERROR);
	    }
	    break;	       
   case OPT_STRONGSUBSUMPTION:
	    StrongUnitForwardSubsumption = true;
	    break;
      case OPT_HEU_INT:
	    PStackPushInt(h_parms->i_parms, CLStateGetIntArg(handle,
							     arg));
	    break;
      case OPT_HEU_FLOAT:
	    DStackPush(h_parms->d_parms, CLStateGetFloatArg(handle,
							    arg));
	    break;
      case OPT_HEU_STRING:
	    PStackPushP(h_parms->s_parms, arg);
	    break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}

void print_help(FILE* out)
{
   fprintf(out, "\n\
\n\
E " VERSION "\n\
\n\
Usage: eprover [options] [files]\n\
\n\
Read sets of clauses and try to refute them.\n\
\n");
   PrintOptions(stdout, opts);
   fprintf(out, "\n\
Copyright 1998 by Stephan Schulz, schulz@informatik.tu-muenchen.de\n\
\n\
You can find the latest version of E and additional information at\n\
http://wwwjessen.informatik.tu-muenchen.de/~schulz/WORK/eprover.html.\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program (it should be contained in the top level\n\
directory of the distribution in the file COPYING); if not, write to\n\
the Free Software Foundation, Inc., 59 Temple Place, Suite 330,\n\
Boston, MA  02111-1307 USA\n\
\n\
The original copyright holder can be contacted as\n\
\n\
Stephan Schulz\n\
Technische Universitaet Muenchen\n\
Fakultaet fuer Informatik\n\
Arcisstrasse 20\n\
D-80290 Muenchen\n\
Germany\n\
");

}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


